@@ -22,6 +22,7 @@
 #include "config.h"
 #include <fcntl.h>
 #include <glib/gi18n-lib.h>
+#include <stdio.h>
 
 #include <blockdev/lvm.h>
 
@@ -39,10 +40,7 @@
 #include "udiskslinuxvolumegroupobject.h"
 
 #include "udiskslvm2daemonutil.h"
-#include "udiskslvm2dbusutil.h"
-#include "udiskslvm2util.h"
 #include "jobhelpers.h"
-#include "udisks-lvm2-generated.h"
 
 /**
  * SECTION:udiskslinuxlogicalvolume
@@ -135,6 +133,107 @@ udisks_linux_logical_volume_new (void)
 
 /* ---------------------------------------------------------------------------------------------------- */
 
+static GVariant *
+build_segment (UDisksDaemon *daemon,
+               UDisksVolumeGroup *group,
+               BDLVMSEGdata *seg)
+{
+  GVariantBuilder seg_builder;
+  UDisksObject *block_object;
+  guint64 extent_size = udisks_volume_group_get_extent_size (group);
+
+  g_variant_builder_init (&seg_builder, G_VARIANT_TYPE ("(tto)"));
+  g_variant_builder_add (&seg_builder, "t", seg->pv_start_pe * extent_size);
+  g_variant_builder_add (&seg_builder, "t", seg->size_pe * extent_size);
+  if (seg->pvdev == NULL)
+    {
+      g_variant_builder_add (&seg_builder, "o", "/");
+    }
+  else
+    {
+      block_object = udisks_daemon_find_block_by_device_file_and_symlinks (daemon, seg->pvdev);
+      if (block_object)
+        {
+          g_variant_builder_add (&seg_builder, "o", g_dbus_object_get_object_path (G_DBUS_OBJECT (block_object)));
+          g_object_unref (block_object);
+        }
+      else
+        g_variant_builder_add (&seg_builder, "o", "/notfound");
+    }
+  return g_variant_builder_end (&seg_builder);
+}
+
+static GVariant *build_structure (UDisksDaemon *daemon, UDisksVolumeGroup *group, BDLVMLVdata *lv, BDLVMLVdata **all);
+
+static gboolean
+lvnameeq (gchar *name, gchar *maybe_bracketed_name)
+{
+  size_t len = strlen (maybe_bracketed_name);
+  return (g_strcmp0 (name, maybe_bracketed_name) == 0
+          || (maybe_bracketed_name[0] == '['
+              && maybe_bracketed_name[len-1] == ']'
+              && strlen (name) == len-2
+              && strncmp (name, maybe_bracketed_name+1, len-2) == 0));
+}
+
+static GVariant *
+build_structures (UDisksDaemon *daemon,
+                  UDisksVolumeGroup *group,
+                  gchar **lv_names,
+                  BDLVMLVdata **all)
+{
+  GVariantBuilder list_builder;
+  g_variant_builder_init (&list_builder, G_VARIANT_TYPE ("aa{sv}"));
+  for (int i = 0; lv_names[i]; i++)
+    {
+      int j;
+      for (j = 0; all[j]; j++) {
+        if (lvnameeq (lv_names[i], all[j]->lv_name))
+          break;
+      }
+      if (all[j])
+        g_variant_builder_add_value (&list_builder, build_structure (daemon, group, all[j], all));
+    }
+  return g_variant_builder_end (&list_builder);
+}
+
+static GVariant *
+build_structure (UDisksDaemon *daemon,
+                 UDisksVolumeGroup *group,
+                 BDLVMLVdata *lv,
+                 BDLVMLVdata **all)
+{
+  GVariantBuilder lv_builder;
+
+  g_variant_builder_init (&lv_builder, G_VARIANT_TYPE_VARDICT);
+  if (lv->lv_name[0] == '[')
+    {
+      gchar *name = g_strndup (lv->lv_name + 1, strlen (lv->lv_name) - 2);
+      g_variant_builder_add (&lv_builder, "{sv}", "name", g_variant_new_string (name));
+      g_free (name);
+    }
+  else
+    g_variant_builder_add (&lv_builder, "{sv}", "name", g_variant_new_string (lv->lv_name));
+
+  g_variant_builder_add (&lv_builder, "{sv}", "type", g_variant_new_string (lv->segtype));
+  g_variant_builder_add (&lv_builder, "{sv}", "size", g_variant_new_uint64 (lv->size));
+
+  if (lv->segs) {
+    GVariantBuilder segs_builder;
+    g_variant_builder_init (&segs_builder, G_VARIANT_TYPE ("a(tto)"));
+    for (int i = 0; lv->segs[i]; i++)
+      g_variant_builder_add_value (&segs_builder, build_segment (daemon, group, lv->segs[i]));
+    g_variant_builder_add (&lv_builder, "{sv}", "segments", g_variant_builder_end (&segs_builder));
+  } else {
+    if (lv->data_lvs)
+      g_variant_builder_add (&lv_builder, "{sv}", "data", build_structures (daemon, group, lv->data_lvs, all));
+    if (lv->metadata_lvs)
+      g_variant_builder_add (&lv_builder, "{sv}", "metadata", build_structures (daemon, group, lv->metadata_lvs, all));
+  }
+
+  return g_variant_builder_end (&lv_builder);
+}
+
 /**
  * udisks_linux_logical_volume_update:
  * @logical_volume: A #UDisksLinuxLogicalVolume.
@@ -148,16 +247,24 @@ udisks_linux_logical_volume_update (UDisksLinuxLogicalVolume     *logical_volume
                                     UDisksLinuxVolumeGroupObject *group_object,
                                     BDLVMLVdata                  *lv_info,
                                     BDLVMLVdata                  *meta_lv_info,
+                                    BDLVMLVdata                 **all_lv_infos,
                                     gboolean                     *needs_polling_ret)
 {
+  UDisksLinuxModuleLVM2 *module;
+  UDisksDaemon *daemon;
   UDisksLogicalVolume *iface;
+  UDisksVolumeGroup *group;
   const char *type;
   gboolean active;
   const char *pool_objpath;
   const char *origin_objpath;
   guint64 size = 0;
 
+  module = udisks_linux_volume_group_object_get_module (group_object);
+  daemon = udisks_module_get_daemon (UDISKS_MODULE (module));
+
   iface = UDISKS_LOGICAL_VOLUME (logical_volume);
+  group = udisks_object_get_volume_group (UDISKS_OBJECT (group_object));
 
   udisks_logical_volume_set_name (iface, lv_info->lv_name);
   udisks_logical_volume_set_uuid (iface, lv_info->uuid);
@@ -176,6 +283,8 @@ udisks_linux_logical_volume_update (UDisksLinuxLogicalVolume     *logical_volume
 
       if (target_type == 't' && volume_type == 't')
         type = "pool";
+      else if (volume_type == 'd')
+        type = "vdopool";
       if (meta_lv_info && meta_lv_info->size)
         size += meta_lv_info->size;
 
@@ -183,6 +292,20 @@ udisks_linux_logical_volume_update (UDisksLinuxLogicalVolume     *logical_volume
         active = TRUE;
     }
   udisks_logical_volume_set_type_ (iface, type);
+  if (g_strcmp0 (lv_info->segtype, "error") == 0)
+    udisks_logical_volume_set_layout (iface, "linear");
+  else
+    udisks_logical_volume_set_layout (iface, lv_info->segtype);
+  if ((g_str_has_prefix (lv_info->segtype, "raid") && g_strcmp0 (lv_info->segtype, "raid0") != 0)
+      || g_strcmp0 (lv_info->segtype, "mirror") == 0)
+    {
+      udisks_logical_volume_set_sync_ratio (iface, lv_info->copy_percent / 100.0);
+      if (lv_info->copy_percent != 100)
+        *needs_polling_ret = TRUE;
+    }
+  else
+    udisks_logical_volume_set_sync_ratio (iface, 1.0);
+
   udisks_logical_volume_set_active (iface, active);
   udisks_logical_volume_set_size (iface, size);
 
@@ -199,7 +322,7 @@ udisks_linux_logical_volume_update (UDisksLinuxLogicalVolume     *logical_volume
   udisks_logical_volume_set_metadata_allocated_ratio (iface, lv_info->metadata_percent / 100.0);
 
   pool_objpath = "/";
-  if (lv_info->pool_lv)
+  if (lv_info->pool_lv && g_strcmp0 (lv_info->segtype, "thin") == 0)
     {
       UDisksLinuxLogicalVolumeObject *pool_object = udisks_linux_volume_group_object_find_logical_volume_object (group_object, lv_info->pool_lv);
       if (pool_object)
@@ -232,23 +355,31 @@ udisks_linux_logical_volume_update (UDisksLinuxLogicalVolume     *logical_volume
       logical_volume->needs_udev_hack = FALSE;
       g_free (dev_file);
     }
+
+  udisks_logical_volume_set_structure (iface, build_structure (daemon, group, lv_info, all_lv_infos));
+
+  g_dbus_interface_skeleton_flush (G_DBUS_INTERFACE_SKELETON (iface));
+  g_clear_object (&group);
 }
 
 void
 udisks_linux_logical_volume_update_etctabs (UDisksLinuxLogicalVolume     *logical_volume,
                                             UDisksLinuxVolumeGroupObject *group_object)
 {
+  UDisksLinuxModuleLVM2 *module;
   UDisksDaemon *daemon;
   UDisksLogicalVolume *iface;
   const gchar *uuid;
 
-  daemon = udisks_linux_volume_group_object_get_daemon (group_object);
+  module = udisks_linux_volume_group_object_get_module (group_object);
+  daemon = udisks_module_get_daemon (UDISKS_MODULE (module));
   iface = UDISKS_LOGICAL_VOLUME (logical_volume);
   uuid = udisks_logical_volume_get_uuid (iface);
 
   udisks_logical_volume_set_child_configuration (iface,
                                                  udisks_linux_find_child_configuration (daemon,
                                                                                         uuid));
+  g_dbus_interface_skeleton_flush (G_DBUS_INTERFACE_SKELETON (iface));
 }
 
 /* ---------------------------------------------------------------------------------------------------- */
@@ -385,6 +516,7 @@ common_setup (UDisksLinuxLogicalVolume           *volume,
               UDisksDaemon                      **daemon,
               uid_t                              *out_uid)
 {
+  UDisksLinuxModuleLVM2 *module;
   gboolean rc = FALSE;
   GError *error = NULL;
 
@@ -395,7 +527,8 @@ common_setup (UDisksLinuxLogicalVolume           *volume,
       goto out;
     }
 
-  *daemon = udisks_linux_logical_volume_object_get_daemon (*object);
+  module = udisks_linux_logical_volume_object_get_module (*object);
+  *daemon = udisks_module_get_daemon (UDISKS_MODULE (module));
 
   if (!udisks_daemon_util_get_caller_uid_sync (*daemon,
                                                invocation,
@@ -403,15 +536,14 @@ common_setup (UDisksLinuxLogicalVolume           *volume,
                                                out_uid,
                                                &error))
     {
-      g_dbus_method_invocation_return_gerror (invocation, error);
-      g_clear_error (&error);
+      g_dbus_method_invocation_take_error (invocation, error);
       goto out;
     }
 
   /* Policy check. */
   UDISKS_DAEMON_CHECK_AUTHORIZATION (*daemon,
                                      UDISKS_OBJECT (*object),
-                                     lvm2_policy_action_id,
+                                     LVM2_POLICY_ACTION_ID,
                                      options,
                                      auth_err_msg,
                                      invocation);
@@ -432,7 +564,7 @@ handle_delete (UDisksLogicalVolume   *_volume,
   uid_t caller_uid;
   gboolean teardown_flag = FALSE;
   UDisksLinuxVolumeGroupObject *group_object;
-  LVJobData data;
+  LVJobData data = {0,};
   struct WaitData wait_data;
 
   g_variant_lookup (options, "tear-down", "b", &teardown_flag);
@@ -483,7 +615,7 @@ handle_delete (UDisksLogicalVolume   *_volume,
                                                          wait_for_logical_volume_object,
                                                          &wait_data,
                                                          NULL,
-                                                         10, /* timeout_seconds */
+                                                         UDISKS_DEFAULT_WAIT_TIMEOUT,
                                                          &error))
     {
       g_prefix_error (&error,
@@ -502,6 +634,65 @@ handle_delete (UDisksLogicalVolume   *_volume,
 
 /* ---------------------------------------------------------------------------------------------------- */
 
+static gboolean
+handle_repair (UDisksLogicalVolume   *_volume,
+               GDBusMethodInvocation *invocation,
+               const gchar *const    *arg_pvs,
+               GVariant              *options)
+{
+  GError *error = NULL;
+  UDisksLinuxLogicalVolume *volume = UDISKS_LINUX_LOGICAL_VOLUME (_volume);
+  UDisksLinuxLogicalVolumeObject *object = NULL;
+  UDisksDaemon *daemon = NULL;
+  uid_t caller_uid;
+  UDisksLinuxVolumeGroupObject *group_object;
+  LVJobData data = {0,};
+  g_auto(GStrv) pvs = NULL;
+
+  if (!common_setup (volume, invocation, options,
+                     N_("Authentication is required to repair a logical volume"),
+                     &object, &daemon, &caller_uid))
+    goto out;
+
+  group_object = udisks_linux_logical_volume_object_get_volume_group (object);
+  data.vg_name = udisks_linux_volume_group_object_get_name (group_object);
+  data.lv_name = udisks_linux_logical_volume_object_get_name (object);
+  pvs = udisks_daemon_util_lvm2_gather_pvs (daemon, group_object, arg_pvs, &error);
+  if (pvs == NULL)
+    {
+      g_dbus_method_invocation_take_error (invocation, error);
+      goto out;
+    }
+  data.new_lv_pvs = (const gchar **)pvs;
+
+  if (!udisks_daemon_launch_threaded_job_sync (daemon,
+                                               UDISKS_OBJECT (object),
+                                               "lvm-lvol-repair",
+                                               caller_uid,
+                                               lvrepair_job_func,
+                                               &data,
+                                               NULL, /* user_data_free_func */
+                                               NULL, /* GCancellable */
+                                               &error))
+    {
+      g_dbus_method_invocation_return_error (invocation,
+                                             UDISKS_ERROR,
+                                             UDISKS_ERROR_FAILED,
+                                             "Error repairing logical volume: %s",
+                                             error->message);
+      g_clear_error (&error);
+      goto out;
+    }
+
+  udisks_logical_volume_complete_repair (_volume, invocation);
+
+ out:
+  g_clear_object (&object);
+  return TRUE;
+}
+
+/* ---------------------------------------------------------------------------------------------------- */
+
 static UDisksObject *
 wait_for_logical_volume_object (UDisksDaemon *daemon,
                                 gpointer      user_data)
@@ -522,18 +713,20 @@ wait_for_logical_volume_path (UDisksLinuxVolumeGroupObject  *group_object,
                               GError                       **error)
 {
   struct WaitData data;
+  UDisksLinuxModuleLVM2 *module;
   UDisksDaemon *daemon;
   UDisksObject *volume_object;
   const gchar *object_path;
 
   data.group_object = group_object;
   data.name = name;
-  daemon = udisks_linux_volume_group_object_get_daemon (group_object);
+  module = udisks_linux_volume_group_object_get_module (group_object);
+  daemon = udisks_module_get_daemon (UDISKS_MODULE (module));
   volume_object = udisks_daemon_wait_for_object_sync (daemon,
                                                       wait_for_logical_volume_object,
                                                       &data,
                                                       NULL,
-                                                      10, /* timeout_seconds */
+                                                      UDISKS_DEFAULT_WAIT_TIMEOUT,
                                                       error);
   if (volume_object == NULL)
     return NULL;
@@ -557,7 +750,7 @@ handle_rename (UDisksLogicalVolume   *_volume,
   uid_t caller_uid;
   UDisksLinuxVolumeGroupObject *group_object;
   const gchar *lv_objpath;
-  LVJobData data;
+  LVJobData data = {0,};
 
   if (!common_setup (volume, invocation, options,
                      N_("Authentication is required to rename a logical volume"),
@@ -619,7 +812,9 @@ handle_resize (UDisksLogicalVolume   *_volume,
   UDisksDaemon *daemon;
   uid_t caller_uid;
   UDisksLinuxVolumeGroupObject *group_object;
-  LVJobData data;
+  LVJobData data = {0,};
+  gchar **opt_pvs = NULL;
+  g_auto(GStrv) pvs = NULL;
 
   if (!common_setup (volume, invocation, options,
                      N_("Authentication is required to resize a logical volume"),
@@ -635,6 +830,19 @@ handle_resize (UDisksLogicalVolume   *_volume,
   data.force = FALSE;
   g_variant_lookup (options, "resize_fsys", "b", &(data.resize_fs));
   g_variant_lookup (options, "force", "b", &(data.force));
+  g_variant_lookup (options, "pvs", "^a&o", &opt_pvs);
+
+  if (opt_pvs)
+    {
+      pvs = udisks_daemon_util_lvm2_gather_pvs (daemon, group_object, (const gchar **)opt_pvs, &error);
+      if (pvs == NULL)
+        {
+          g_dbus_method_invocation_take_error (invocation, error);
+          goto out;
+        }
+    }
+
+  data.new_lv_pvs = (const gchar **)pvs;
 
   if (!udisks_daemon_launch_threaded_job_sync (daemon,
                                                UDISKS_OBJECT (object),
@@ -659,6 +867,7 @@ handle_resize (UDisksLogicalVolume   *_volume,
 
  out:
   g_clear_object (&object);
+  g_free (opt_pvs);
   return TRUE;
 }
 
@@ -713,7 +922,7 @@ handle_activate (UDisksLogicalVolume *_volume,
   uid_t caller_uid;
   UDisksLinuxVolumeGroupObject *group_object;
   UDisksObject *block_object = NULL;
-  LVJobData data;
+  LVJobData data = {0,};
 
   if (!common_setup (volume, invocation, options,
                      N_("Authentication is required to activate a logical volume"),
@@ -748,7 +957,7 @@ handle_activate (UDisksLogicalVolume *_volume,
                                                      wait_for_logical_volume_block_object,
                                                      object,
                                                      NULL,
-                                                     10, /* timeout_seconds */
+                                                     UDISKS_DEFAULT_WAIT_TIMEOUT,
                                                      &error);
   if (block_object == NULL)
     {
@@ -781,7 +990,7 @@ handle_deactivate (UDisksLogicalVolume   *_volume,
   UDisksDaemon *daemon;
   uid_t caller_uid;
   UDisksLinuxVolumeGroupObject *group_object;
-  LVJobData data;
+  LVJobData data = {0,};
 
   if (!common_setup (volume, invocation, options,
                      N_("Authentication is required to deactivate a logical volume"),
@@ -815,7 +1024,7 @@ handle_deactivate (UDisksLogicalVolume   *_volume,
                                                          wait_for_logical_volume_block_object,
                                                          object,
                                                          NULL,
-                                                         10, /* timeout_seconds */
+                                                         UDISKS_DEFAULT_WAIT_TIMEOUT,
                                                          &error))
     {
       g_prefix_error (&error,
@@ -848,7 +1057,7 @@ handle_create_snapshot (UDisksLogicalVolume   *_volume,
   uid_t caller_uid;
   UDisksLinuxVolumeGroupObject *group_object;
   const gchar *lv_objpath = NULL;
-  LVJobData data;
+  LVJobData data = {0,};
 
   if (!common_setup (volume, invocation, options,
                      N_("Authentication is required to create a snapshot of a logical volume"),
@@ -903,23 +1112,13 @@ handle_cache_attach (UDisksLogicalVolume   *volume_,
                      const gchar           *cache_name,
                      GVariant              *options)
 {
-#ifndef HAVE_LVMCACHE
-
-  g_dbus_method_invocation_return_error (invocation,
-                                         UDISKS_ERROR,
-                                         UDISKS_ERROR_FAILED,
-                                         N_("LVMCache not enabled at compile time."));
-  return TRUE;
-
-#else
-
   GError *error = NULL;
   UDisksLinuxLogicalVolume *volume = UDISKS_LINUX_LOGICAL_VOLUME (volume_);
   UDisksLinuxLogicalVolumeObject *object = NULL;
   UDisksDaemon *daemon;
   uid_t caller_uid;
   UDisksLinuxVolumeGroupObject *group_object;
-  LVJobData data;
+  LVJobData data = {0,};
 
   if (!common_setup (volume, invocation, options,
                      N_("Authentication is required to convert logical volume to cache"),
@@ -955,34 +1154,22 @@ out:
   g_clear_object (&object);
 
   return TRUE;
-
-#endif /* HAVE_LVMCACHE */
 }
 
 
 static gboolean
-handle_cache_detach_or_split (UDisksLogicalVolume  *volume_,
+handle_cache_detach_or_split (UDisksLogicalVolume    *volume_,
                               GDBusMethodInvocation  *invocation,
                               GVariant               *options,
-                              gboolean               destroy)
+                              gboolean                destroy)
 {
-#ifndef HAVE_LVMCACHE
-
-  g_dbus_method_invocation_return_error (invocation,
-                                         UDISKS_ERROR,
-                                         UDISKS_ERROR_FAILED,
-                                         N_("LVMCache not enabled at compile time."));
-  return TRUE;
-
-#else
-
   GError *error = NULL;
   UDisksLinuxLogicalVolume *volume = UDISKS_LINUX_LOGICAL_VOLUME (volume_);
   UDisksLinuxLogicalVolumeObject *object = NULL;
   UDisksDaemon *daemon;
   uid_t caller_uid;
   UDisksLinuxVolumeGroupObject *group_object;
-  LVJobData data;
+  LVJobData data = {0,};
 
   if (!common_setup (volume, invocation, options,
                      N_("Authentication is required to split cache pool LV off of a cache LV"),
@@ -1018,8 +1205,6 @@ out:
   g_clear_object (&object);
 
   return TRUE;
-
-#endif /* HAVE_LVMCACHE */
 }
 
 static gboolean
@@ -1046,6 +1231,7 @@ logical_volume_iface_init (UDisksLogicalVolumeIface *iface)
   iface->handle_delete = handle_delete;
   iface->handle_rename = handle_rename;
   iface->handle_resize = handle_resize;
+  iface->handle_repair = handle_repair;
   iface->handle_activate = handle_activate;
   iface->handle_deactivate = handle_deactivate;
   iface->handle_create_snapshot = handle_create_snapshot;
