@@ -1,12 +1,9 @@
 import dbus
 import os
 import re
-import six
 import shutil
 import tarfile
 import tempfile
-import time
-import unittest
 import configparser
 
 import gi
@@ -36,6 +33,13 @@ def _get_blkid_version():
         raise RuntimeError('Failed to determine blkid version from: %s' % out)
     return Version(m.groups()[0])
 
+def _get_luks_version(disk):
+    ret, out = udiskstestcase.UdisksTestCase.run_command("cryptsetup luksDump %s" % disk)
+    m = re.search(r'Version:\s*([1-2])', out)
+    if not m or len(m.groups()) != 1:
+        raise RuntimeError('Failed to determine LUKS version of device: %s.' % disk)
+    return int(m.groups()[0])
+
 
 class UdisksEncryptedTest(udiskstestcase.UdisksTestCase):
     '''This is an encrypted device test suite'''
@@ -156,13 +160,13 @@ class UdisksEncryptedTest(udiskstestcase.UdisksTestCase):
 
         # no password
         msg = 'org.freedesktop.UDisks2.Error.Failed: No key available.*'
-        with six.assertRaisesRegex(self, dbus.exceptions.DBusException, msg):
+        with self.assertRaisesRegex(dbus.exceptions.DBusException, msg):
             disk.Unlock("", self.no_options,
                         dbus_interface=self.iface_prefix + '.Encrypted')
 
         # wrong password
         msg = 'org.freedesktop.UDisks2.Error.Failed: Error unlocking %s *' % self.vdevs[0]
-        with six.assertRaisesRegex(self, dbus.exceptions.DBusException, msg):
+        with self.assertRaisesRegex(dbus.exceptions.DBusException, msg):
             disk.Unlock('abcdefghijklmn', self.no_options,
                         dbus_interface=self.iface_prefix + '.Encrypted')
 
@@ -304,7 +308,7 @@ class UdisksEncryptedTest(udiskstestcase.UdisksTestCase):
 
         # should not be possible to close mounted luks
         msg = 'org.freedesktop.UDisks2.Error.Failed: Error locking'
-        with six.assertRaisesRegex(self, dbus.exceptions.DBusException, msg):
+        with self.assertRaisesRegex(dbus.exceptions.DBusException, msg):
             disk.Lock(self.no_options, dbus_interface=self.iface_prefix + '.Encrypted')
 
         # now unmount it and try to close it again
@@ -326,7 +330,7 @@ class UdisksEncryptedTest(udiskstestcase.UdisksTestCase):
 
         # old password, should fail
         msg = 'org.freedesktop.UDisks2.Error.Failed: Error unlocking %s *' % self.vdevs[0]
-        with six.assertRaisesRegex(self, dbus.exceptions.DBusException, msg):
+        with self.assertRaisesRegex(dbus.exceptions.DBusException, msg):
             disk.Unlock(self.PASSPHRASE, self.no_options,
                         dbus_interface=self.iface_prefix + '.Encrypted')
 
@@ -452,6 +456,24 @@ class UdisksEncryptedTestLUKS1(UdisksEncryptedTest):
         # offset value is in 512B blocks; we need to multiply to get the real metadata size
         return  int(m.group(1)) * 512
 
+    def test_convert(self):
+        disk = self.vdevs[0]
+        device = self.get_device(disk)
+        self._create_luks(device, self.PASSPHRASE)
+        self.assertEqual(1, _get_luks_version(disk))
+
+        self.addCleanup(self._remove_luks, device)
+        self.udev_settle()
+        device.Lock(self.no_options, dbus_interface=self.iface_prefix + '.Encrypted')
+
+        device.Convert("luks2", self.no_options,
+                       dbus_interface=self.iface_prefix + '.Encrypted')
+        self.assertEqual(2, _get_luks_version(disk))
+
+        device.Convert("luks1", self.no_options,
+                       dbus_interface=self.iface_prefix + '.Encrypted')
+        self.assertEqual(1, _get_luks_version(disk))
+
 class UdisksEncryptedTestLUKS2(UdisksEncryptedTest):
     '''This is a LUKS2 encrypted device test suite'''
 
@@ -531,7 +553,7 @@ class UdisksEncryptedTestLUKS2(UdisksEncryptedTest):
         # kernel keyring support and no passphrase for LUKS 2 given = fail
         if self._get_key_location('/dev/' + clear_dev) == 'keyring':
             msg = 'org.freedesktop.UDisks2.Error.Failed: Error resizing encrypted device /dev/dm-[0-9]+: Insufficient (permissions|persmissions) to resize device. *'
-            with six.assertRaisesRegex(self, dbus.exceptions.DBusException, msg):
+            with self.assertRaisesRegex(dbus.exceptions.DBusException, msg):
                 device.Resize(dbus.UInt64(100*1024*1024), self.no_options,
                               dbus_interface=self.iface_prefix + '.Encrypted')
 
@@ -540,7 +562,7 @@ class UdisksEncryptedTestLUKS2(UdisksEncryptedTest):
         d['passphrase'] = 'wrongpassphrase'
         msg = 'org.freedesktop.UDisks2.Error.Failed: Error resizing encrypted device /dev/dm-[0-9]+: '\
               'Failed to activate device: (Operation not permitted|Incorrect passphrase)'
-        with six.assertRaisesRegex(self, dbus.exceptions.DBusException, msg):
+        with self.assertRaisesRegex(dbus.exceptions.DBusException, msg):
             device.Resize(dbus.UInt64(100*1024*1024), d,
                           dbus_interface=self.iface_prefix + '.Encrypted')
 
@@ -562,6 +584,77 @@ class UdisksEncryptedTestLUKS2(UdisksEncryptedTest):
         clear_size3 = self.get_block_size(clear_dev)
         self.assertEqual(clear_size3, clear_size)
 
+    def test_convert_xfail(self):
+        disk = self.vdevs[0]
+        device = self.get_device(disk)
+        self._create_luks(device, self.PASSPHRASE)
+        self.assertEqual(2, _get_luks_version(disk))
+
+        self.addCleanup(self._remove_luks, device)
+        self.udev_settle()
+        device.Lock(self.no_options, dbus_interface=self.iface_prefix + '.Encrypted')
+
+        # check that the device is not LUKS1 compatible
+        ret, out = udiskstestcase.UdisksTestCase.run_command("cryptsetup luksDump %s" % disk)
+        m = re.search(r'PBKDF:\s*(.*)', out)
+        if not m or len(m.groups()) != 1:
+            raise RuntimeError('Failed to determine PBKDF of LUKS device: %s.' % disk)
+        self.assertEqual("argon2id", m.groups()[0])
+
+        # check that conversion fails
+        msg = 'org.freedesktop.UDisks2.Error.Failed: Error converting encrypted device /dev/.+: Conversion failed: Invalid argument'
+        with self.assertRaisesRegex(dbus.exceptions.DBusException, msg):
+            device.Convert("luks1", self.no_options,
+                           dbus_interface=self.iface_prefix + '.Encrypted')
+
+        self.assertEqual(2, _get_luks_version(disk))
+
+    def test_header_backup_restore(self):
+        disk = self.vdevs[0]
+        device = self.get_device(disk)
+        self._create_luks(device, self.PASSPHRASE)
+
+        self.addCleanup(self._remove_luks, device)
+        self.udev_settle()
+
+        # check that backup normally works
+        with tempfile.TemporaryDirectory() as temp_dir:
+            backup_file_path = os.path.join(temp_dir, "udisks_encrypted_header_backup.luks")
+
+            device.HeaderBackup(backup_file_path, self.no_options,
+                                dbus_interface=self.iface_prefix + '.Encrypted')
+            self.assertTrue(os.path.exists(backup_file_path))
+
+            # check that backup will fail if there is already a file under the specified path
+            msg = 'org.freedesktop.UDisks2.Error.Failed: Error backing up header of encrypted device /dev/.+: Failed to backup LUKS header: Invalid argument'
+            with self.assertRaisesRegex(dbus.exceptions.DBusException, msg):
+                device.HeaderBackup(backup_file_path, self.no_options,
+                                    dbus_interface=self.iface_prefix + '.Encrypted')
+
+            self.assertTrue(os.path.exists(backup_file_path))
+
+            # check that after reaping device and restoring header, cryptsetup will recognize header
+            device.Lock(self.no_options, dbus_interface=self.iface_prefix + '.Encrypted')
+            DISK_SIZE = 100 # magic number from `targetcli_config.json`
+            ret, out = udiskstestcase.UdisksTestCase.run_command("dd if=/dev/zero of=%s bs=1M count=%d" % (disk, DISK_SIZE))
+            self.assertEqual(ret, 0)
+            ret, out = udiskstestcase.UdisksTestCase.run_command("cryptsetup luksDump %s" % disk)
+            self.assertEqual(1, ret)
+            self.assertTrue(("Device %s is not a valid LUKS device." % disk) in out)
+            # wait for changes to propagate from udev to udisks
+            fstype = self.get_property(device, '.Block', 'IdType')
+            fstype.assertEqual('') # this method waits
+
+            device = self.get_device(disk)
+            device.RestoreEncryptedHeader(backup_file_path, self.no_options,
+                                 dbus_interface=self.iface_prefix + '.Block')
+            ret, out = udiskstestcase.UdisksTestCase.run_command("cryptsetup luksDump %s" % disk)
+            self.assertEqual(ret, 0)
+
+        # get the Encrypted interface back
+        device = self.get_device(disk)
+        self.assertHasIface(device, "org.freedesktop.UDisks2.Encrypted")
+
     def _get_default_luks_version(self):
         manager = self.get_object('/Manager')
         default_encryption_type = self.get_property(manager, '.Manager', 'DefaultEncryptionType')
@@ -644,6 +737,93 @@ class UdisksEncryptedTestLUKS2(UdisksEncryptedTest):
         dbus_cleartext = self.get_property(device, '.Encrypted', 'CleartextDevice')
         dbus_cleartext.assertEqual('/')
 
+    @udiskstestcase.tag_test(udiskstestcase.TestTags.UNSAFE)
+    def test_teardown_locked(self):
+        # based on https://github.com/storaged-project/udisks/issues/1204
+        crypttab = self.read_file('/etc/crypttab')
+        self.addCleanup(self.write_file, '/etc/crypttab', crypttab)
+        fstab = self.read_file('/etc/fstab')
+        self.addCleanup(self.write_file, '/etc/fstab', fstab)
+
+        disk_name = os.path.basename(self.vdevs[0])
+        disk = self.get_object('/block_devices/' + disk_name)
+
+        disk.Format('gpt', self.no_options, dbus_interface=self.iface_prefix + '.Block')
+
+        # create LUKS and ext4 filesystem
+        options = dbus.Dictionary(signature='sv')
+        options['encrypt.passphrase'] = self.PASSPHRASE
+        options['encrypt.type'] = 'luks2'
+        options['tear-down'] = dbus.Boolean(True)
+
+        conf_fstab = dbus.Dictionary({'dir': self.str_to_ay('fakemountpoint'),
+                                      'type': self.str_to_ay('ext4'),
+                                      'opts': self.str_to_ay('noauto,nofail'),
+                                      'freq': 0, 'passno': 0,
+                                      'track-parents': True},
+                                     signature=dbus.Signature('sv'))
+        conf_crypttab = dbus.Dictionary({'name': self.str_to_ay('udisks_luks_teardown'),
+                                         'passphrase-contents': self.str_to_ay(self.PASSPHRASE),
+                                         'options': self.str_to_ay('noauto,nofail'),
+                                         'track-parents': True},
+                                         signature=dbus.Signature('sv'))
+        options['config-items'] = dbus.Array([('fstab', conf_fstab), ('crypttab', conf_crypttab)])
+
+        path = disk.CreatePartitionAndFormat(dbus.UInt64(1024**2), dbus.UInt64(100 * 1024**2), '', '',
+                                             self.no_options,'ext4', options,
+                                             dbus_interface=self.iface_prefix + '.PartitionTable')
+        self.addCleanup(self.remove_file, '/etc/luks-keys/udisks_luks_teardown', ignore_nonexistent=True)
+        part = self.bus.get_object(self.iface_prefix, path)
+        self.assertIsNotNone(part)
+        self.udev_settle()
+
+        # Now lock the device (without that step, everything works)
+        luks_path = self.get_property(part, '.Encrypted', 'CleartextDevice')
+        luks_dev = self.bus.get_object(self.iface_prefix, luks_path.value)
+        self.assertIsNotNone(luks_dev)
+        luks_uuid = self.get_property_raw(luks_dev, '.Block', 'IdUUID')
+        self.assertIsNotNone(luks_uuid)
+        part.Lock(self.no_options, dbus_interface=self.iface_prefix + '.Encrypted')
+
+        # Try to reformat the drive
+        options = dbus.Dictionary(signature='sv')
+        options['tear-down'] = dbus.Boolean(True)
+        disk.Format('gpt', options, dbus_interface=self.iface_prefix + '.Block')
+
+        # Check that crypttab and fstab records are gone
+        crypttab = self.read_file('/etc/crypttab')
+        self.assertNotIn('udisks_luks_teardown', crypttab)
+
+        fstab = self.read_file('/etc/fstab')
+        self.assertNotIn(luks_uuid, fstab)
+
+    def test_create_pbkdf_extra(self):
+        disk_name = os.path.basename(self.vdevs[0])
+        disk = self.get_object('/block_devices/' + disk_name)
+
+        # create LUKS without specifying version
+        options = dbus.Dictionary(signature='sv')
+        options['encrypt.passphrase'] = self.PASSPHRASE
+        options['encrypt.pbkdf'] = 'pbkdf2'
+        options['encrypt.iterations'] = dbus.UInt32(10000)
+
+        disk.Format('ext4', options,
+                    dbus_interface=self.iface_prefix + '.Block')
+
+        self.addCleanup(self._remove_luks, disk)
+        self.udev_settle()
+
+        _ret, out = self.run_command("cryptsetup luksDump %s" % self.vdevs[0])
+        m = re.search(r"PBKDF:\s*(\S+)\s*", out)
+        if not m or len(m.groups()) != 1:
+            self.fail("Failed to get pbkdf information from:\n%s" % out)
+        self.assertEqual(m.group(1), "pbkdf2")
+
+        m = re.search(r"Iterations:\s*(\S+)\s*", out)
+        if not m or len(m.groups()) != 1:
+            self.fail("Failed to get pbkdf information from:\n%s" % out)
+        self.assertEqual(m.group(1), "10000")
+
 
 class UdisksEncryptedTestBITLK(udiskstestcase.UdisksTestCase):
 
@@ -726,5 +906,21 @@ class UdisksEncryptedTestBITLK(udiskstestcase.UdisksTestCase):
         objects = udisks.GetManagedObjects(dbus_interface='org.freedesktop.DBus.ObjectManager')
         self.assertNotIn(str(crypt_path), objects.keys())
 
+        # unlock with keyfile contents
+        d = dbus.Dictionary(signature='sv')
+        d['keyfile_contents'] = self.str_to_ay(self.passphrase, False)
+        crypt_path = self.loop.Unlock("", d,
+                                      dbus_interface=self.iface_prefix + '.Encrypted')
+        self.assertIsNotNone(crypt_path)
+        crypt_dev = self.bus.get_object(self.iface_prefix, crypt_path)
+        self.assertIsNotNone(crypt_dev)
+
+        dbus_cleartext = self.get_property(self.loop, '.Encrypted', 'CleartextDevice')
+        dbus_cleartext.assertEqual(str(crypt_path))
+        dbus_type = self.get_property(self.loop, '.Encrypted', 'HintEncryptionType')
+        dbus_type.assertEqual("BITLK")
+
+        self.loop.Lock(self.no_options, dbus_interface=self.iface_prefix + '.Encrypted')
+
 
 del UdisksEncryptedTest  # skip UdisksEncryptedTest
