@@ -368,7 +368,9 @@ static const struct
   { "ID_DRIVE_FLASH_SD", "flash_sd", FALSE, TRUE },
   { "ID_DRIVE_FLASH_SDHC", "flash_sdhc", FALSE, TRUE },
   { "ID_DRIVE_FLASH_SDXC", "flash_sdxc", FALSE, TRUE },
-  { "ID_DRIVE_FLASH_MMC", "flash_mmc", FALSE, TRUE },
+  { "ID_DRIVE_FLASH_SDIO", "flash_sdio", FALSE, TRUE },
+  { "ID_DRIVE_FLASH_SD_COMBO", "flash_sd_combo", FALSE, TRUE },
+  { "ID_DRIVE_FLASH_MMC", "flash_mmc", TRUE, FALSE },
   { "ID_DRIVE_FLOPPY", "floppy", FALSE, TRUE },
   { "ID_DRIVE_FLOPPY_ZIP", "floppy_zip", FALSE, TRUE },
   { "ID_DRIVE_FLOPPY_JAZ", "floppy_jaz", FALSE, TRUE },
@@ -408,6 +410,7 @@ static const struct
   { "ID_DRIVE_MEDIA_FLASH_SD", "flash_sd" },
   { "ID_DRIVE_MEDIA_FLASH_SDHC", "flash_sdhc" },
   { "ID_DRIVE_MEDIA_FLASH_SDXC", "flash_sdxc" },
+  { "ID_DRIVE_MEDIA_FLASH_SDIO", "flash_sdio" },
   { "ID_DRIVE_MEDIA_FLASH_MMC", "flash_mmc" },
   { "ID_DRIVE_MEDIA_FLOPPY", "floppy" },
   { "ID_DRIVE_MEDIA_FLOPPY_ZIP", "floppy_zip" },
@@ -442,6 +445,34 @@ ptr_str_array_compare (const gchar **a,
   return g_strcmp0 (*a, *b);
 }
 
+static gboolean
+find_removable (GUdevDevice *device)
+{
+  GUdevDevice *parent;
+
+  parent = g_object_ref (device);
+  while (parent)
+  {
+    GUdevDevice *d;
+    const gchar * const *removable;
+    removable = g_udev_device_get_sysfs_attr_as_strv (parent, "removable");
+    if (removable && removable[0])
+    {
+      if (g_strcmp0 (removable[0], "removable") == 0)
+      {
+        g_object_unref (parent);
+        return TRUE;
+      }
+    }
+
+    d = parent;
+    parent = g_udev_device_get_parent (d);
+    g_object_unref (d);
+  }
+
+  return FALSE;
+}
+
 static void
 set_media (UDisksDrive       *iface,
            UDisksLinuxDevice *device,
@@ -476,7 +507,16 @@ set_media (UDisksDrive       *iface,
   g_ptr_array_sort (media_compat_array, (GCompareFunc) ptr_str_array_compare);
   g_ptr_array_add (media_compat_array, NULL);
 
-  removable = ejectable = g_udev_device_get_sysfs_attr_as_boolean (device->udev_device, "removable");
+  if (udisks_linux_device_subsystem_is_nvme (device))
+    {
+      removable = find_removable (device->udev_device);
+      ejectable = FALSE;
+    }
+  else
+    {
+      removable = ejectable = g_udev_device_get_sysfs_attr_as_boolean (device->udev_device, "removable");
+    }
+
   if (force_non_removable)
     removable = FALSE;
   if (force_removable)
@@ -484,6 +524,9 @@ set_media (UDisksDrive       *iface,
   udisks_drive_set_media_removable (iface, removable);
   if (is_pc_floppy_drive)
     ejectable = FALSE;
+  /* MMC/SD: disrespect the sysfs 'removable' attr as it usually doesn't reflect reality */
+  if (g_str_has_prefix (g_udev_device_get_name (device->udev_device), "mmcblk"))
+    ejectable = removable;
   udisks_drive_set_ejectable (iface, ejectable);
 
   media_in_drive = NULL;
@@ -540,7 +583,9 @@ set_rotation_rate (UDisksDrive       *iface,
   else
     {
       rate = -1;
-      if (device->ata_identify_device_data != NULL)
+      if (g_udev_device_has_property (device->udev_device, "ID_ATA_ROTATION_RATE_RPM"))
+        rate = g_udev_device_get_property_as_int (device->udev_device, "ID_ATA_ROTATION_RATE_RPM");
+      else if (device->ata_identify_device_data != NULL)
         {
           guint word_217 = 0;
 
@@ -707,9 +752,9 @@ udisks_linux_drive_update (UDisksLinuxDrive       *drive,
   UDisksDrive *iface = UDISKS_DRIVE (drive);
   UDisksLinuxDevice *device = NULL;
   const gchar *serial = NULL;
-  guint64 size;
-  gboolean media_available;
-  gboolean media_change_detected;
+  guint64 size = 0;
+  gboolean media_available = TRUE;
+  gboolean media_change_detected = TRUE;
   gboolean is_pc_floppy_drive = FALSE;
   gboolean removable_hint = FALSE;
   UDisksDaemon *daemon;
@@ -794,6 +839,57 @@ udisks_linux_drive_update (UDisksLinuxDrive       *drive,
        *  - lookup Revision from fwrev and hwrev in sysfs
        */
     }
+  else if (udisks_linux_device_subsystem_is_nvme (device))
+    {
+      gchar *model = NULL;
+      gchar *firmware = NULL;
+
+      serial = g_udev_device_get_sysfs_attr (device->udev_device, "serial");
+      model = g_strdup (g_udev_device_get_sysfs_attr (device->udev_device, "model"));
+      firmware = g_strdup (g_udev_device_get_sysfs_attr (device->udev_device, "firmware_rev"));
+
+      /* Since we issue the 'NVMe Identify Controller' command to the drive which
+       * is an actual ioctl and the drive may be busy or in a temporary offline state,
+       * leaving us with missing data, it's better to leave old information around
+       * instead of clearing to blank.
+       */
+      if (device->nvme_ctrl_info)
+        {
+          /* TODO: decode nvme_ctrl_info's Vendor ID */
+          /* TODO: rotation rate - add support for NVMe Log Page – Rotational Media Information Log to libblockdev first */
+          if (model == NULL || strlen (model) == 0)
+            {
+              g_free (model);
+              model = g_strdup (device->nvme_ctrl_info->model_number);
+            }
+          if (firmware == NULL || strlen (firmware) == 0)
+            {
+              g_free (firmware);
+              firmware = g_strdup (device->nvme_ctrl_info->firmware_ver);
+            }
+          if (device->nvme_ctrl_info->serial_number && strlen (device->nvme_ctrl_info->serial_number) > 0)
+            serial = device->nvme_ctrl_info->serial_number;
+          size = device->nvme_ctrl_info->size_total;
+          /* WWN is namespace specific, leaving unset */
+        }
+      else
+        {
+          udisks_debug ("Missing probed NVMe controller information for %s",
+                        g_udev_device_get_sysfs_path (device->udev_device));
+        }
+      if (model)
+        {
+          g_strstrip (model);
+          udisks_drive_set_model (iface, model);
+          g_free (model);
+        }
+      if (firmware)
+        {
+          g_strstrip (firmware);
+          udisks_drive_set_revision (iface, firmware);
+          g_free (firmware);
+        }
+    }
   else
     {
       const gchar *vendor;
@@ -877,9 +973,10 @@ udisks_linux_drive_update (UDisksLinuxDrive       *drive,
   udisks_drive_set_serial (iface, serial);
 
   /* common bits go here */
-  size = udisks_daemon_util_block_get_size (device->udev_device,
-                                            &media_available,
-                                            &media_change_detected);
+  if (size == 0)
+    size = udisks_daemon_util_block_get_size (device->udev_device,
+                                              &media_available,
+                                              &media_change_detected);
   udisks_drive_set_size (iface, size);
   udisks_drive_set_media_available (iface, media_available);
   udisks_drive_set_media_change_detected (iface, media_change_detected);
@@ -889,7 +986,6 @@ udisks_linux_drive_update (UDisksLinuxDrive       *drive,
 
   if (udisks_drive_get_media_removable (iface) ||
       g_strcmp0 (udisks_drive_get_connection_bus (iface), "usb") == 0 ||
-      g_strcmp0 (udisks_drive_get_connection_bus (iface), "sdio") == 0 ||
       g_strcmp0 (udisks_drive_get_connection_bus (iface), "ieee1394") == 0)
     removable_hint = TRUE;
   udisks_drive_set_removable (iface, removable_hint);
@@ -1181,7 +1277,7 @@ handle_set_configuration (UDisksDrive           *_drive,
   if (!udisks_daemon_util_file_set_contents (path,
                                              data,
                                              data_len,
-                                             0600, /* mode to use if non-existant */
+                                             0600, /* mode to use if non-existent */
                                              &error))
     {
       g_dbus_method_invocation_take_error (invocation, error);
@@ -1536,7 +1632,7 @@ handle_power_off (UDisksDrive           *_drive,
     }
   fd = -1;
 
-  device = udisks_linux_drive_object_get_device (object, TRUE /* get_hw */);
+  device = udisks_linux_drive_object_get_device (object, FALSE /* get_hw */);
   if (device == NULL)
     {
       g_dbus_method_invocation_return_error (invocation,
