@@ -162,15 +162,14 @@ update_metadata_size (UDisksLinuxEncrypted   *encrypted,
                       UDisksLinuxBlockObject *object)
 {
   UDisksLinuxDevice *device;
-  guint64 metadata_size;
+  BDCryptoLUKSInfo *info = NULL;
   GError *error = NULL;
 
   device = udisks_linux_block_object_get_device (object);
 
-  metadata_size = bd_crypto_luks_get_metadata_size (g_udev_device_get_device_file (device->udev_device),
-                                                    &error);
-
-  if (error != NULL)
+  info = bd_crypto_luks_info (g_udev_device_get_device_file (device->udev_device),
+                              &error);
+  if (!info)
     {
       udisks_warning ("Error getting '%s' metadata_size: %s (%s, %d)",
                       g_udev_device_get_device_file (device->udev_device),
@@ -178,10 +177,13 @@ update_metadata_size (UDisksLinuxEncrypted   *encrypted,
                       g_quark_to_string (error->domain),
                       error->code);
       g_clear_error (&error);
+      udisks_encrypted_set_metadata_size (UDISKS_ENCRYPTED (encrypted), 0);
     }
+  else
+    udisks_encrypted_set_metadata_size (UDISKS_ENCRYPTED (encrypted), info->metadata_size);
 
   g_object_unref (device);
-  udisks_encrypted_set_metadata_size (UDISKS_ENCRYPTED (encrypted), metadata_size);
+  bd_crypto_luks_info_free (info);
 }
 
 static void
@@ -463,8 +465,8 @@ handle_unlock (UDisksEncrypted        *encrypted,
       goto out;
     }
 
-  /* fallback mechanism: keyfile_contents (for LUKS) -> passphrase -> crypttab_passphrase -> TCRYPT keyfiles -> error (no key) */
-  if (is_luks && udisks_variant_lookup_binary (options, "keyfile_contents", &effective_passphrase))
+  /* fallback mechanism: keyfile_contents (for LUKS and BITLK) -> passphrase -> crypttab_passphrase -> TCRYPT keyfiles -> error (no key) */
+  if ((is_luks || is_bitlk) && udisks_variant_lookup_binary (options, "keyfile_contents", &effective_passphrase))
     {
       /* effective_passphrase was set to keyfile_contents, nothing more to do here */
     }
@@ -783,7 +785,7 @@ udisks_linux_encrypted_lock (UDisksLinuxEncrypted   *encrypted,
     }
 
   device = udisks_linux_block_object_get_device (UDISKS_LINUX_BLOCK_OBJECT (cleartext_object));
-  data.map_name = g_udev_device_get_sysfs_attr (device->udev_device, "dm/name");
+  data.map_name = (gchar *) g_udev_device_get_sysfs_attr (device->udev_device, "dm/name");
 
   if (is_luks)
     close_func = luks_close_job_func;
@@ -1036,11 +1038,12 @@ handle_resize (UDisksEncrypted       *encrypted,
   GError *error = NULL;
   UDisksBaseJob *job = NULL;
   GString *effective_passphrase = NULL;
+  BDCryptoKeyslotContext *context = NULL;
 
   object = udisks_daemon_util_dup_object (encrypted, &error);
   if (object == NULL)
     {
-      g_dbus_method_invocation_take_error (invocation, error);
+      g_dbus_method_invocation_return_gerror (invocation, error);
       goto out;
     }
 
@@ -1066,7 +1069,6 @@ handle_resize (UDisksEncrypted       *encrypted,
   if (!udisks_daemon_util_get_caller_uid_sync (daemon, invocation, NULL /* GCancellable */, &caller_uid, &error))
     {
       g_dbus_method_invocation_return_gerror (invocation, error);
-      g_clear_error (&error);
       goto out;
     }
 
@@ -1117,6 +1119,30 @@ handle_resize (UDisksEncrypted       *encrypted,
                                                      invocation))
     goto out;
 
+  if (udisks_variant_lookup_binary (options, "keyfile_contents", &effective_passphrase))
+    ;
+  else if (udisks_variant_lookup_binary (options, "passphrase", &effective_passphrase))
+    ;
+  else
+    effective_passphrase = NULL;
+
+  if (effective_passphrase)
+    {
+      context = bd_crypto_keyslot_context_new_passphrase ((const guint8 *) effective_passphrase->str,
+                                                          effective_passphrase->len,
+                                                          &error);
+      if (!context)
+        {
+          g_dbus_method_invocation_return_error (invocation,
+                                             UDISKS_ERROR,
+                                             UDISKS_ERROR_FAILED,
+                                             "Error resizing encrypted device %s: %s",
+                                             udisks_block_get_device (cleartext_block),
+                                             error->message);
+          goto out;
+        }
+    }
+
   job = udisks_daemon_launch_simple_job (daemon,
                                          UDISKS_OBJECT (object),
                                          "encrypted-resize",
@@ -1129,21 +1155,13 @@ handle_resize (UDisksEncrypted       *encrypted,
       goto out;
     }
 
-  if (udisks_variant_lookup_binary (options, "keyfile_contents", &effective_passphrase))
-    ;
-  else if (udisks_variant_lookup_binary (options, "passphrase", &effective_passphrase))
-    ;
-  else
-    effective_passphrase = NULL;
-
   udisks_linux_block_encrypted_lock (block);
 
   /* TODO: implement progress parsing for udisks_job_set_progress(_valid) */
-  if (! bd_crypto_luks_resize_luks2_blob (udisks_block_get_device (cleartext_block),
-                                          size / 512,
-                                          effective_passphrase ? (const guint8*) effective_passphrase->str : NULL,
-                                          effective_passphrase ? effective_passphrase->len : 0,
-                                          &error))
+  if (! bd_crypto_luks_resize (udisks_block_get_device (cleartext_block),
+                               size / 512,
+                               context,
+                               &error))
     {
       g_dbus_method_invocation_return_error (invocation,
                                              UDISKS_ERROR,
@@ -1170,11 +1188,239 @@ handle_resize (UDisksEncrypted       *encrypted,
   g_clear_object (&object);
   g_clear_error (&error);
   udisks_string_wipe_and_free (effective_passphrase);
+  bd_crypto_keyslot_context_free (context);
   return TRUE; /* returning TRUE means that we handled the method invocation */
 }
 
 /* ---------------------------------------------------------------------------------------------------- */
 
+/* runs in thread dedicated to handling method call */
+static gboolean
+handle_convert (UDisksEncrypted       *encrypted,
+                GDBusMethodInvocation *invocation,
+                const gchar           *target_version,
+                GVariant              *options)
+{
+  UDisksObject *object = NULL;
+  UDisksBlock *block;
+  UDisksDaemon *daemon;
+  UDisksState *state = NULL;
+  uid_t caller_uid;
+  const gchar *action_id = NULL;
+  const gchar *message = NULL;
+  GError *error = NULL;
+  UDisksBaseJob *job = NULL;
+  BDCryptoLUKSVersion bd_target_version;
+
+  object = udisks_daemon_util_dup_object (encrypted, &error);
+  if (object == NULL)
+    {
+      g_dbus_method_invocation_return_gerror (invocation, error);
+      goto out;
+    }
+
+  block = udisks_object_peek_block (object);
+  daemon = udisks_linux_block_object_get_daemon (UDISKS_LINUX_BLOCK_OBJECT (object));
+  state = udisks_daemon_get_state (daemon);
+
+  udisks_linux_block_object_lock_for_cleanup (UDISKS_LINUX_BLOCK_OBJECT (object));
+  udisks_state_check_block (state, udisks_linux_block_object_get_device_number (UDISKS_LINUX_BLOCK_OBJECT (object)));
+
+  /* Fail if the device is not a LUKS device */
+  if (!(g_strcmp0 (udisks_block_get_id_usage (block), "crypto") == 0 &&
+        g_strcmp0 (udisks_block_get_id_type (block), "crypto_LUKS") == 0))
+    {
+      g_dbus_method_invocation_return_error (invocation,
+                                             UDISKS_ERROR,
+                                             UDISKS_ERROR_FAILED,
+                                             "Device %s does not appear to be a LUKS device",
+                                             udisks_block_get_device (block));
+      goto out;
+    }
+
+  if (!udisks_daemon_util_get_caller_uid_sync (daemon, invocation, NULL /* GCancellable */, &caller_uid, &error))
+    {
+      g_dbus_method_invocation_return_gerror (invocation, error);
+      goto out;
+    }
+
+  action_id = "org.freedesktop.udisks2.modify-device";
+  /* Translators: Shown in authentication dialog when the user
+   * requests conversion of an encrypted block device.
+   *
+   * Do not translate $(drive), it's a placeholder and
+   * will be replaced by the name of the drive/device in question
+   */
+  message = N_("Authentication is required to convert device $(drive) to a different LUKS version.");
+  if (! udisks_daemon_util_setup_by_user (daemon, object, caller_uid))
+    {
+      if (udisks_block_get_hint_system (block))
+        {
+          action_id = "org.freedesktop.udisks2.modify-device-system";
+        }
+      else if (! udisks_daemon_util_on_user_seat (daemon, UDISKS_OBJECT (object), caller_uid))
+        {
+          action_id = "org.freedesktop.udisks2.modify-device-other-seat";
+        }
+    }
+
+  /* Check that the user is actually authorized to convert the device. */
+  if (! udisks_daemon_util_check_authorization_sync (daemon,
+                                                     object,
+                                                     action_id,
+                                                     options,
+                                                     message,
+                                                     invocation))
+    goto out;
+
+  if (g_strcmp0 (target_version, "luks1") == 0) {
+      bd_target_version = BD_CRYPTO_LUKS_VERSION_LUKS1;
+    } else if (g_strcmp0 (target_version, "luks2") == 0) {
+      bd_target_version = BD_CRYPTO_LUKS_VERSION_LUKS2;
+    } else {
+      g_dbus_method_invocation_return_error (invocation, UDISKS_ERROR, UDISKS_ERROR_FAILED,
+                                             "Unsupported target LUKS version: '%s'. Only 'luks1' and 'luks2' are supported.",
+                                             target_version);
+      goto out;
+    }
+
+  job = udisks_daemon_launch_simple_job (daemon,
+                                         UDISKS_OBJECT (object),
+                                         "encrypted-convert",
+                                         caller_uid,
+                                         NULL);
+  if (job == NULL)
+    {
+      g_dbus_method_invocation_return_error (invocation, UDISKS_ERROR, UDISKS_ERROR_FAILED,
+                                             "Failed to create a job object");
+      goto out;
+    }
+
+  udisks_linux_block_encrypted_lock (block);
+
+  if (! bd_crypto_luks_convert (udisks_block_get_device (block),
+                                bd_target_version,
+                                &error))
+    {
+      g_dbus_method_invocation_return_error (invocation,
+                                             UDISKS_ERROR,
+                                             UDISKS_ERROR_FAILED,
+                                             "Error converting encrypted device %s: %s",
+                                             udisks_block_get_device (block),
+                                             error->message);
+      udisks_simple_job_complete (UDISKS_SIMPLE_JOB (job), FALSE, error->message);
+      udisks_linux_block_encrypted_unlock (block);
+      goto out;
+    }
+
+  udisks_linux_block_encrypted_unlock (block);
+
+  udisks_encrypted_complete_convert (encrypted, invocation);
+  udisks_simple_job_complete (UDISKS_SIMPLE_JOB (job), TRUE, NULL);
+
+ out:
+  if (object != NULL)
+    udisks_linux_block_object_release_cleanup_lock (UDISKS_LINUX_BLOCK_OBJECT (object));
+  if (state != NULL)
+    udisks_state_check (state);
+  g_clear_object (&object);
+  g_clear_error (&error);
+  return TRUE; /* returning TRUE means that we handled the method invocation */
+}
+
+/* ---------------------------------------------------------------------------------------------------- */
+
+/* runs in thread dedicated to handling method call */
+static gboolean
+handle_header_backup (UDisksEncrypted       *encrypted,
+                      GDBusMethodInvocation *invocation,
+                      const gchar           *backup_file,
+                      GVariant              *options)
+{
+    UDisksObject *object = NULL;
+    UDisksBlock *block;
+    UDisksDaemon *daemon;
+    UDisksState *state = NULL;
+    uid_t caller_uid;
+    GError *error = NULL;
+    UDisksBaseJob *job = NULL;
+
+    object = udisks_daemon_util_dup_object (encrypted, &error);
+    if (object == NULL)
+      {
+        g_dbus_method_invocation_return_gerror (invocation, error);
+        goto out;
+      }
+
+    block = udisks_object_peek_block (object);
+    daemon = udisks_linux_block_object_get_daemon (UDISKS_LINUX_BLOCK_OBJECT (object));
+    state = udisks_daemon_get_state (daemon);
+
+    udisks_linux_block_object_lock_for_cleanup (UDISKS_LINUX_BLOCK_OBJECT (object));
+    udisks_state_check_block (state, udisks_linux_block_object_get_device_number (UDISKS_LINUX_BLOCK_OBJECT (object)));
+
+    /* Fail if the device is not a LUKS device */
+    if (!(g_strcmp0 (udisks_block_get_id_usage (block), "crypto") == 0 &&
+          g_strcmp0 (udisks_block_get_id_type (block), "crypto_LUKS") == 0))
+      {
+        g_dbus_method_invocation_return_error (invocation,
+                                               UDISKS_ERROR,
+                                               UDISKS_ERROR_FAILED,
+                                               "Device %s does not appear to be a LUKS device",
+                                               udisks_block_get_device (block));
+        goto out;
+      }
+
+    if (!udisks_daemon_util_get_caller_uid_sync (daemon, invocation, NULL /* GCancellable */, &caller_uid, &error))
+      {
+        g_dbus_method_invocation_return_gerror (invocation, error);
+        goto out;
+      }
+
+    job = udisks_daemon_launch_simple_job (daemon,
+                                           UDISKS_OBJECT (object),
+                                           "encrypted-header-backup",
+                                           caller_uid,
+                                           NULL);
+    if (job == NULL)
+      {
+        g_dbus_method_invocation_return_error (invocation, UDISKS_ERROR, UDISKS_ERROR_FAILED,
+                                               "Failed to create a job object");
+        goto out;
+      }
+
+    udisks_linux_block_encrypted_lock (block);
+
+    if (!bd_crypto_luks_header_backup (udisks_block_get_device (block), backup_file, &error))
+      {
+        g_dbus_method_invocation_return_error (invocation,
+                                               UDISKS_ERROR,
+                                               UDISKS_ERROR_FAILED,
+                                               "Error backing up header of encrypted device %s: %s",
+                                               udisks_block_get_device (block),
+                                               error->message);
+        udisks_simple_job_complete (UDISKS_SIMPLE_JOB (job), FALSE, error->message);
+        udisks_linux_block_encrypted_unlock (block);
+        goto out;
+      }
+
+    udisks_linux_block_encrypted_unlock (block);
+
+    udisks_encrypted_complete_header_backup (encrypted, invocation);
+    udisks_simple_job_complete (UDISKS_SIMPLE_JOB (job), TRUE, NULL);
+
+    out:
+    if (object != NULL)
+        udisks_linux_block_object_release_cleanup_lock (UDISKS_LINUX_BLOCK_OBJECT (object));
+    if (state != NULL)
+        udisks_state_check (state);
+    g_clear_object (&object);
+    g_clear_error (&error);
+    return TRUE; /* returning TRUE means that we handled the method invocation */
+}
+
+/* ---------------------------------------------------------------------------------------------------- */
+
 static void
 encrypted_iface_init (UDisksEncryptedIface *iface)
 {
@@ -1182,4 +1428,6 @@ encrypted_iface_init (UDisksEncryptedIface *iface)
   iface->handle_lock                = handle_lock;
   iface->handle_change_passphrase   = handle_change_passphrase;
   iface->handle_resize              = handle_resize;
+  iface->handle_convert             = handle_convert;
+  iface->handle_header_backup       = handle_header_backup;
 }
